/*

  Test of height mapping
  Started: 06/05/94
  Author: Leo Sutic - Watcom conversion by Lary Myers
  Module: HTX.C (Original module: RAYCAST.CPP then HT.C)

  Authors:

    Leo Sutic
      Original Borland C++ code

    Lary Myers
      Initial conversion to WATCOM C++ 9.5 with some speedups

    Joe Slayton
      Variable size and posn viewport
      Variable terrain sizes, including sizes larger than 320x200
      Faster draw routine (leads to fuzzier drawing)
      Horizontal tilting (tends to scale as it tilts, good for low angles)
      Viewing angle independent of viewport width
      Broke one source file into several (might be good or bad)
      Renamed as HTX.C

*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <dos.h>
#include <math.h>
#include <time.h>
#include "loadpcx.h"
#include "kbd.h"
#include "vid.h"
#include "angle.h"

/* pixel dimensions of physical screen for video mode */
#define XSCRNMAX 320 /* pixel width of viewing window */
#define YSCRNMAX 200 /* height of viewing window */

/* max length of a single ray */
#define MAXRAYLENGTH 200

/* max width of viewport (max # rays to cast) */
#define VPXMAX 320

/* maximum width of the color/height maps */
#define MAXWORLDX 1024

/* viewport definition (not allowed to be larger than physical screen
   dimensions and not allowed to go off the screen) */
typedef struct {

  int x, y;           /* starting location on screen (pixels) */
  int w, h;           /* width and height of viewport (pixels) */

  int HFOV;           /* horizontal field of view (degrees) */
  int VFOV;           /* vertical field of view (map units) */

  int RayLength;      /* length of rays cast in this viewport */

  int Double;         /* double-up the columns (twice as fast, 1/2 res) */

  int X;              /* user X position */
  int Y;              /* user Y position */
  int A;              /* user view angle */
  int Horizon;        /* elevation of horizon (for pitch of helicopter) */
  int Alt;            /* current user altitude above pixel */
  int Speed;          /* forward speed of user */
  int Tilt;           /* tilt across the horizon (pixels) */

  long dAngle;        /* step for columns to get HFOV angle */

  unsigned char *Buf;            /* data buffer (w*h bytes) */
  unsigned char *Line[YSCRNMAX]; /* addr of each line screen buffer [Buf] */

  int  RayStart[MAXRAYLENGTH];   /* RayOrg + Horizon */
  long RangeTable[MAXRAYLENGTH]; /* perspective transform along ray */

  /* delta from current user posn to each pixel in visible surface */
  char XDelta[VPXMAX][MAXRAYLENGTH];
  char YDelta[VPXMAX][MAXRAYLENGTH];

  /* 0-based origin of ray adjusted for perspective */
  int RayOrg[MAXRAYLENGTH];

} VPType;

/* hardware video memory buffer */
unsigned char *VidBuf = (unsigned char *) 0xA0000;

/* definition for maps (similar to PCX structure) */
typedef struct {
  int w, h;     /* width and height of map */
  unsigned char *Map;   /* data */
} MapType;

MapType Height;   /* altitude bitmap for world */
MapType Texture;  /* color bitmap for world */
MapType Sky;      /* bitmap for sky */

/* offset into texture and height map for each line in the world */
int WorldInd[MAXWORLDX];

/*-------------------------------------------------------------------------*/

int VPPaintLine(VPType *VP,
             long Column, /* current column on screen */
             long Angle,  /* current angle for ray */
             long VOfs)   /* vertical offset into line */
/*
  Traces along a line from the current position to a point on a circle some
  distance (RAYLENGTH) away in world space at angle Angle.  Applies a simple
  transform to each element of the line to determine the perspective
  transform, then converts the point to screen coords.
  A line is then drawn from the old screen Y posn to the current Y posn using
  the color of the texture map at that point.
  Returns the highest point found on the line.
*/
{


  /* offset into color and height maps for each point on ray */
  int Ofs[MAXRAYLENGTH];

  /* loop counter for ray cast */
  int RayPos;

  {
  /* unit step in X and Y */
  FIXED dx = icos[Angle];
  FIXED dy = isin[Angle];

  /* current position in world */
  FIXED x  = IFIX(VP->X);
  FIXED y  = IFIX(VP->Y);

  /* wrap the rays at the world max X and Y values */
  FIXED MaxX = IFIX(Height.w);
  FIXED MaxY = IFIX(Height.h);

  /* scratch pointer to offset array, used to fill it */
  int *o = Ofs;

  /* go through the ray, calculating offsets for each point */
  /* we calculate from the current point outward to reduce the effects of
     cumulative errors in teh fixed-point arithmetic.  The errors appeared
     as "grass" in the displayed image, caused by a roundoff error in the
     choice of points closest to the origin. */
  for (o = Ofs, RayPos = VP->RayLength - 1; RayPos > 0; --RayPos) {

    /* make sure the current XY posn is still within the world */
    if (x >= MaxX) x -= MaxX; else if (x < 0) x += MaxX;
    if (y >= MaxY) y -= MaxY; else if (y < 0) y += MaxY;

    /* determine offset into height and color maps */
    *o++ = WorldInd[UNIFIX(y)] + UNIFIX(x);

    /* adjust x and y */
    x += dx;
    y += dy;

  } /* for */

  }


  {

  /* value of previous height calculation */
  int OldPos = VP->h;

  /* highest point found (return value of function) */
  int MaxPt = VP->h;

  /* step through the whole length of the ray, adjusting the X and Y posns */
  for (RayPos = VP->RayLength - 1; RayPos > 0; --RayPos) {

    int Offset; /* offset into world and texture maps */
    int High; /* height value (height map, then adjust for perspective) */

    /* determine offset into height and color maps */
    Offset = Ofs[RayPos];

    /* determine user height above ground at this point */
    High = Height.Map[Offset] - VP->Alt;

    /* adjust the height for the distance (multiply by ~1/dist) */
    High = (High * VP->RangeTable[RayPos]) >> FIXSHIFT;

    /* determine where we are relative to a unit-height line at this pt */
    High = VP->RayStart[RayPos] - High;

    /* adjust for tilt parm */
    High += VOfs;

    /* make sure we stay onscreen */
    if (High > VP->h)
      High = VP->h;
    else if (High < 0)
      High = 0;

    /* Are we beyond the last height and in bounds for the screen? */
    if (High > OldPos) {

      /* pointer to screen buf for fill */
      unsigned char *d;

      /* texture map value for fill */
      int Color = Texture.Map[Offset];

      /* figure how much we have to draw from the old position */
      int NumRows = High - OldPos;

      /* make sure the number of rows is correct */
      if (NumRows > 0 && NumRows < VP->h) {

        /* position to start filling */
        int BegRow = OldPos;

        /* make sure we don't overflow off the bottom of the screen */
        if (BegRow + NumRows > VP->h)
          NumRows = VP->h - BegRow;

        /* insert the data into the viewport buffer */
        d = VP->Line[BegRow] + Column;
        while (NumRows-- > 0) {
          *d = Color;
          d += VP->w;
        } /* while */

      } /* if */

      /* check high point, saving higher */
      if (High < MaxPt)
        MaxPt = High;

    } /* if */

    /* we're higher now, so remember it */
    OldPos = High;

  } /* for */

  /* return highest point found (may exceed viewport size) */
  return MaxPt;


  }

} /* VPPaintLine */

/*-------------------------------------------------------------------------*/

void VPPaintBuf(VPType *VP)
/*
  Draws the current viewport into VP->Buf.  Displays background, then
  places the elevation data on screen in front.
*/
{

  FIXED dTilt = 2 * VP->Tilt / VP->w; /* change in the tilt per column */
  FIXED Tilt  = -VP->Tilt;            /* current tilt */

#if 0
  // Draw the sky in 2 sections, first start out with the column of the sky
  // for our current angle and draw as much as possible, then start with column
  // zero of the sky and draw any remaining columns
  {
  short Angle;
  short i1, i2;       /* scratch */

  /* determine where the sky background will start */
  Angle = (VP->A % Height.w) % VP->w;

  /* determine how much to draw in 1st pass */
  if ((i1 = VP->w - Angle) != 0)
    DrawSkyRange(VP, Angle, 0, i1, VP->h/*, UNIFIX(Tilt)*/);

  /* figure amount left to finish */
  if ((i2 = VP->w - i1) != 0)
    DrawSkyRange(VP, 0, i1, i2, VP->h/*, UNIFIX(Tilt)*/);

  }

#endif

  /* draw each column onscreen, starting at the user angle. the angle for
     each column is bumped one, wrapping around to 0 if we exceed the max
     angle. */
  {
  long Column;  /* current column on screen */
  FIXED Angle;  /* current angle */

  /* determine starting angle (half a screen back from the current angle) */
  Angle = IFIX(VP->A) - (VP->w / 2) * VP->dAngle;
  if (Angle < 0)
    Angle += IFIX(MAXANGLE);

  /* reset tilt angle */
  Tilt = -VP->Tilt;

  /* draw each column */
  for (Column = 0; Column < VP->w; Column++) {

    /* if it's time to draw a ray */
    if (!VP->Double || (VP->Double && ((Column & 1) != 0))) {

      /* determine angle and tilt values used locally */
      int a = UNIFIX(Angle);
      int t = UNIFIX(Tilt);

      /* draw ground (remember highest point found) */
      int High = VPPaintLine(VP, Column, a, t);

      /* if there's sky showing */
      if (High > 0 && High < VP->h) {

        /* copy data from the sky map to the viewport buffer */
        unsigned char *SkyPtr = Sky.Map +
                                ((Sky.h / 2 - t) * Sky.w) + /* y */
                                (a % Sky.w);                /* x */

        /*determine corresponding position in viewport buffer  */
        unsigned char *BufPtr = VP->Buf + /*(High * VP->w)*/ + Column;

        /* copy pixels until we're done */
        do {

          *BufPtr = *SkyPtr;

          BufPtr += VP->w;
          SkyPtr += Sky.w;

        } while (--High != 0);

      } /* if */

    } /* if */
    else {

      /* duplicate the last column */
      unsigned char *s = VP->Buf + Column - 1;
      unsigned char *d = s + 1;
      int i = VP->h;

      do {
        *d = *s;
        d += VP->w;
        s += VP->w;
      } while (--i != 0);

    } /* else */

    /* advance angle and wrap at MAXANGLE */
    if ((Angle += VP->dAngle) >= IFIX(MAXANGLE))
      Angle -= IFIX(MAXANGLE);

    /* advance the Tilt variable */
    Tilt += dTilt;

  } /* for */
  }

} /* VPPaintBuf */

/*-------------------------------------------------------------------------*/

void VPShow(VPType *VP)
/*
  Blit the viewport buffer into video memory.
*/
{

  int i, w;

  /* construct destination pointer */
  unsigned char *d = VidBuf + (XSCRNMAX * VP->y) + VP->x;

  /* construct source pointer */
  unsigned char *s = VP->Buf;

  /* copy each line from the viewpoint buffer to the screen */
  w = VP->w;
  for (i = VP->h; i > 0; --i) {
    memcpy(d, s, w);
    d += XSCRNMAX;
    s += w;
  } /* for */

} /* VPShow */

/*-------------------------------------------------------------------------*/

void VPCalcCameraTables(VPType *VP)
/*
  Calculate data required for the camera given
*/
{

  int i;

  /* init start position of rays */
  for (i = 0; i < VP->RayLength; i++) {
    VP->RayStart[i] = VP->RayOrg[i] + VP->Horizon;
  } /* for */

} /* VPCalcCameraTables */

/*-------------------------------------------------------------------------*/

void VPCalcTables(VPType *VP)
/*
  Calculates data associated with the viewport definition and.
*/
{

  int i;

  /* init start position of rays */
  VP->RayOrg[0] = 500;
  for (i = 1; i < VP->RayLength; i++) {
    VP->RayOrg[i] = VP->RayOrg[0] / i;
  } /* for */

  /* init the address of each entry in the screen buffer line addr table */
  for (i = 0; i < VP->h; i++) {
    VP->Line[i] = &VP->Buf[i * VP->w];
  } /* for */

  /* calculate the scaling vs. range (perspective) factors */
  VP->RangeTable[0] = FFIX(16.0);
  for (i = 1; i < VP->RayLength; i++) {
    VP->RangeTable[i] = (VP->RangeTable[0] / i) * (VP->h / (float) VP->VFOV);
  } /* for */

  /* amount to step angle across screen */
  VP->dAngle = FFIX((VP->HFOV / 360.0) * (MAXANGLE / VP->w));

  /* recalc camera tables */
  VPCalcCameraTables(VP);

} /* VPCalcTables */

/*-------------------------------------------------------------------------*/

void VPInit(VPType *VP)
/*
  Init the global tables that need to be calculated only once per viewpoint.
*/
{

  /* init user view parms */
  VP->X       = 0;         /* corner of world */
  VP->Y       = 0;
  VP->A       = 0;         /* looking north */
  VP->Horizon = VP->h / 2; /* horizon at "normal" attitude (halfway) */
  VP->Alt     = 260;       /* reasonable altitude (above landscape, anyway) */
  VP->Speed   = 3;         /* reasonable speed */
  VP->Tilt    = 0;         /* current tilt of system */

  /* set up viewport */
  VP->Buf = malloc(VP->w * VP->h);
  memset(VP->Buf, 0, VP->w * VP->h);

  /* calculate look-up tables */
  VPCalcTables(VP);

} /* VPInit */

/*-------------------------------------------------------------------------*/

/* definitions for data files */
struct {
  char    FName[32]; /* name of file */
  MapType *Buf;      /* place to store pointer to buffer */
} DataFiles[] = {
  { "SKY.PCX",      &Sky     },
  { "ALTITUDE.PCX", &Height  },
  { "COLOR.PCX",    &Texture },
  { "",             NULL     }
};

/* VGA palette read from last file */
struct VgaPalette PalBuf[256];

short LoadFiles(void)
/*
  Allocate room for and read all of the graphics files involved in the
  program (backdrop, land heights, land color, sky map).
*/
{

  int i;
  int Err = 0; /* error code */

  /* load all files */
  for (i = 0; DataFiles[i].Buf != NULL; i++) {

    struct PcxPix PCX;

    /* read definition for PCX read routine, giving errors as appropriate */
    if (load_pcx(DataFiles[i].FName, &PCX, PalBuf) != pcx_ok) {

      printf("Unable to read file \"%s\"\n", DataFiles[i].FName);

      Err = 1;

      DataFiles[i].Buf->w = 0;
      DataFiles[i].Buf->h = 0;
      DataFiles[i].Buf->Map = NULL;

    } /* if */
    else {

      DataFiles[i].Buf->w   = PCX.width;
      DataFiles[i].Buf->h   = PCX.height;
      DataFiles[i].Buf->Map = PCX.image;

    } /* else */

  } /* for */

  /* got here, no error */
  return Err;

} /* LoadFiles */

/*-------------------------------------------------------------------------*/

void FreeFiles(void)
/*
  Frees all data files in the list DataFiles that have been read.
*/
{

  int i;

  /* free all files allocated */
  for (i = 0; DataFiles[i].Buf != NULL; i++) {

    /* free buffer if something was allocated */
    if (DataFiles[i].Buf->Map != NULL) {
      free(DataFiles[i].Buf->Map);
    } /* if */

  } /* for */

} /* FreeFiles */

/*-------------------------------------------------------------------------*/

void UpdateCamera(VPType *VP)
/*
  Update the keyboard for one pass.
*/
{

  int SpinAngle = MAXANGLE >> 7; /* amount to spin at each command */

  /* move forward */
  if (KbdKeyDown[UP_ARROW_KEY]) {
    int Angle = (VP->A                  ) % MAXANGLE;
    VP->X = (VP->X + ((icos[Angle] * VP->Speed) >> FIXSHIFT));
    VP->Y = (VP->Y + ((isin[Angle] * VP->Speed) >> FIXSHIFT));
    if (VP->X >= Height.w) VP->X -= Height.w;
    else if (VP->X < 0)    VP->X += Height.w;
    if (VP->Y >= Height.h) VP->Y -= Height.h;
    else if (VP->Y < 0)    VP->Y += Height.h;
  } /* if */

  /* move backward */
  if (KbdKeyDown[DOWN_ARROW_KEY]) {
    int Angle = (VP->A + (MAXANGLE / 2)) % MAXANGLE;
    VP->X = (VP->X + ((icos[Angle] * VP->Speed) >> FIXSHIFT));
    VP->Y = (VP->Y + ((isin[Angle] * VP->Speed) >> FIXSHIFT));
    if (VP->X >= Height.w) VP->X -= Height.w;
    else if (VP->X < 0)    VP->X += Height.w;
    if (VP->Y >= Height.h) VP->Y -= Height.h;
    else if (VP->Y < 0)    VP->Y += Height.h;
  } /* if */

  /* rotate left */
  if (KbdKeyDown[LEFT_ARROW_KEY]) {
    VP->A -= SpinAngle;
    if (VP->A < 0)
      VP->A += MAXANGLE;
  } /* if */

  /* rotate right */
  if (KbdKeyDown[RIGHT_ARROW_KEY]) {
    VP->A += SpinAngle;
    if (VP->A >= MAXANGLE)
      VP->A -= MAXANGLE;
  } /* if */

  /* move down if not at bottom */
  if (KbdKeyDown[MINUS_KEY]) {
    if (VP->Alt - VP->Speed > 4)
      VP->Alt -= VP->Speed;
  } /* if */

  /* move up if not topped out */
  if (KbdKeyDown[PLUS_KEY]) {
    if (VP->Alt + VP->Speed < 512)
      VP->Alt += VP->Speed;
  } /* if */

  /* process a single-press keystroke */
  switch (KbdASCII) {

    /* change forward speed */
    case '1' : case '2' : case '3' : case '4' : case '5' :
    case '6' : case '7' : case '8' : case '9' :
      VP->Speed = KbdASCII - '0';
      break;

    case ',' :
      {
      int Angle = (VP->A + 3 * MAXANGLE / 4) % MAXANGLE;
      VP->X = (VP->X + ((icos[Angle] * VP->Speed) >> FIXSHIFT));
      VP->Y = (VP->Y + ((isin[Angle] * VP->Speed) >> FIXSHIFT));
      if (VP->X >= Height.w) VP->X -= Height.w;
      else if (VP->X < 0)    VP->X += Height.w;
      if (VP->Y >= Height.h) VP->Y -= Height.h;
      else if (VP->Y < 0)    VP->Y += Height.h;
      }
      break;

    case '.' :
      {
      int Angle = (VP->A + MAXANGLE / 4) % MAXANGLE;
      VP->X = (VP->X + ((icos[Angle] * VP->Speed) >> FIXSHIFT));
      VP->Y = (VP->Y + ((isin[Angle] * VP->Speed) >> FIXSHIFT));
      if (VP->X >= Height.w) VP->X -= Height.w;
      else if (VP->X < 0)    VP->X += Height.w;
      if (VP->Y >= Height.h) VP->Y -= Height.h;
      else if (VP->Y < 0)    VP->Y += Height.h;
      }
      break;

    /* adjust downward look angle */
    case 201 : /* PGUP */
      if (VP->Horizon > 4) {
        VP->Horizon--;
        VPCalcCameraTables(VP);
      } /* if */
      break;
    case 209 : /* PGDN */
      if (VP->Horizon < 255) {
        VP->Horizon++;
        VPCalcCameraTables(VP);
      } /* if */
      break;

    case '`' :
      {
      int i;
      for (i = 0; i < 200; i++)
        memcpy(VidBuf + 320 * i, Sky.Map + Sky.w * i, 320);

      KbdASCII = 0;
      while (KbdASCII == 0) ;

      }
      break;

    /* adjust left/right tilt */
    case 'A' :
      VP->Tilt += FFIX(1.5);
      break;
    case 'S' :
      VP->Tilt -= FFIX(1.5);
      break;

    /* single or double columns? */
    case 'D' :
      VP->Double ^= 1;
      break;

    /* adjust ray length? */
    case '-' :
      if (VP->RayLength > 1) {
        VP->RayLength--;
        VPCalcTables(VP);
      } /* if */
      break;
    case '=' :
      if (VP->RayLength < MAXRAYLENGTH - 1) {
        VP->RayLength++;
        VPCalcTables(VP);
      } /* if */
      break;

    /* adjust viewing angle? */
    case '[' :
      if (VP->HFOV > 1) {
        VP->HFOV--;
        VPCalcTables(VP);
      } /* if */
      break;
    case ']' :
      if (VP->HFOV < 4096) {
        VP->HFOV++;
        VPCalcTables(VP);
      } /* if */
      break;

    /* adjust viewing vertical scale? */
    case ';' :
      if (VP->VFOV > 1) {
        VP->VFOV--;
        VPCalcTables(VP);
      } /* if */
      break;
    case '\'' :
      if (VP->VFOV < 32767) {
        VP->VFOV++;
        VPCalcTables(VP);
      } /* if */
      break;

  } /* switch */

} /* UpdateCamera */

/*-------------------------------------------------------------------------*/

/* viewport definition */
VPType VP0 = {
   10,  25,     /* start posn */
  300, 150,     /* size (h size = # rays to cast) */
   60, 255,     /* FOV and vert scale */
  XSCRNMAX / 2, /* ray length */
  1             /* double rays */
};

main(short argc, char *argv[])
{

  time_t StartTime; /* starting time of execution */
  int Frames;       /* number of frames drawn */

  int i;            /* loop counter */

  /* allocate and load the files for world, colors, backdrop, etc */
  if (LoadFiles() != 0) {
    return 1;
  } /* if */

  /* calc sin and cos tables */
  InitAngleTable();

  /* init the address of each entry in the world buffer line addr table */
  for (i = 0; i < Height.h; i++) {
    WorldInd[i] = i * Height.w;
  } /* for */

  /* set up the various lookup tables */
  VPInit(&VP0);

  /* set 320x200 video mode */
  SetVideoMode(0x13);

  /* set palette used by height map and other maps */
  SetPalette((unsigned char *) PalBuf);

  /* attach the int-driven keyboard handlers */
  KbdInit();

  /* start counting */
  Frames = 0;
  StartTime = time(NULL);

  do {

    /* flush keyboard buffer */
    KbdASCII = 0;

    /* draw screen */
    VPPaintBuf(&VP0);

    /* blit buffer to screen */
    VPShow(&VP0);

    /* update camera info based on screen */
    UpdateCamera(&VP0);

    /* one more frame */
    Frames++;

  /* until ESC */
  } while (KbdASCII != 27);

  /* get delta time */
  StartTime = time(NULL) - StartTime;

  /* remove the int-driven keyboard handler */
  KbdClose();

  /* return to text mode */
  SetVideoMode(0x03);

#ifndef __386__
  /* prevent Borland C++ from arguing about parameters */
  if (argc == (int) argv) ;
#endif

  /* tell user how many frames / sec */
  if (StartTime > 0) {
    printf("%0.2f Frames/Second\n", Frames / (float) StartTime);
  } /* if */

  /* no bad things happened */
  return 0;

} /* main */

/*=========================================================================*/

